#!/usr/bin/python
"""
Notepaq - Notepad program for Familiar

Copyright (C) 2001 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

ToDo:
1)  Rewrite database to use shelve?
    Currently, we rewrite the whole thing whenever stuff changes.
    With shelve/dbm, we would only rewrite changed stuff, I think.
    This would be better for the FLASH RAM.

"""

try: import pylau
except ImportError: pass

import gtk
import pickle
import os
import sys
import string

import traceback

# - Simple text note base-class
class Note:
    def __init__(self):
        self.modified = 0
        self.text = ""
    def __setstate__(self, state):
        self.modified = 0
        self.text = state
    def __getstate__(self): return self.text
    def set_manager(self, manager):
        self.mgr = manager
    def dump(self):
        if self.mgr: self.mgr.dump()
    def dump_now(self):
        if self.mgr: self.mgr.dump_now()
    def destroy(self):
        if self.mgr: self.mgr.note_remove(self)

# - Manager for list of simple text notes
class NoteManager:
    def __init__(self, filename):
        self.timer = None
        self.modified = 0
        self.filename = filename
        self.notes = []
    def __setstate__(self, state):
        self.modified = 0
        self.timer = None
        self.notes = state
        for note in self.notes:
            note.set_manager(self)
    def __getstate__(self): return self.notes
    def note_add(self, note):
        if note not in self.notes:
            self.modified = 1
            note.set_manager(self)
            self.notes.append(note)
    def note_remove(self, note):
        if note in self.notes:
            self.modified = 1
            self.notes.remove(note)
    def dump(self):
        if not self.timer:
            # - Start the commit timer
            self.timer = gtk.timeout_add(30000, self.on_dump_timeout)
    def dump_now(self):
        dump = self.modified
        self.modified = 0
        for note in self.notes:
            if note.modified:
                note.modified = 0
                dump = 1
        if dump:
            print "notepaq:committed database"
            pickle.dump(self, open(self.filename, "w"))
        return dump
    def on_dump_timeout(self):
        done = self.dump_now()
        if not done: self.timer = None
        return done

# - Loader (will be superceded by familiar framework)
def load_note_manager(filename):
    try: file = open(filename, "r")
    except: NoteManager(filename)
    try:
        manager = pickle.load(file)
        manager.filename = filename
        file.close()
        return manager
    except:
        print "WARNING: error", sys.exc_info()[0], sys.exc_info()[1], "while loading", filename
        traceback.print_tb(sys.exc_info()[2])
        return NoteManager(filename)

class Button(gtk.GtkButton):
    def __init__(self, title, onclicked, name=None):
        gtk.GtkButton.__init__(self,title)
        self.connect("clicked", onclicked)
        if name: self.set_name(name)

class HButtonBox(gtk.GtkHButtonBox):
    def __init__(self, buttons):
        gtk.GtkHButtonBox.__init__(self)
        self.set_spacing(0)
        self.set_child_ipadding_default(0, 0)
        self.set_child_size_default(0,0)
        for button in buttons:
            self.add(button)

# - View a single text note in its own window
class NoteView(gtk.GtkWindow):
    def __init__(self, note, mgrview, title):
        gtk.GtkWindow.__init__(self)
        if title:
            self.set_title(title)
        else:
            self.set_title(mgrview.NoteTitle(note))
        
        self.note = note
        self.mgrview = mgrview

        # - Window & edit box
        textbox = gtk.GtkText()
        textbox.set_word_wrap(1)
        textbox.set_line_wrap(1)
        textbox.insert_defaults(note.text)
        textbox.set_editable(1)
        textbox.set_position(len(note.text))
        self.textbox = textbox

        # - Buttons
        save = Button("Save", self.OnSaveClicked, "Commit")
        cancel = Button("Cancel", self.OnCancelClicked, "Discard")
        buttons = HButtonBox([save, cancel])

        # - Pack the window
        vbox = gtk.GtkVBox()
        vbox.pack_start(textbox)
        vbox.pack_end(buttons, 0, 0)

        self.add(vbox)
        self.show_all()
        self.textbox.grab_focus()

    def OnSaveClicked(self, widget):
        if self.note.text != self.textbox.get_chars(0,-1):
            self.note.text = self.textbox.get_chars(0,-1)
            self.note.modified = 1
        self.mgrview.mgr.note_add(self.note)
        self.mgrview.AddNote(self.note)
        self.note.dump()
        self.destroy()

    def OnCancelClicked(self, widget):
        self.destroy()

class NoteManagerView:
    def __init__(self, manager):
        self.mgr = manager

        # - Window
        window = gtk.GtkWindow()
        window.set_title("Notepaq")
        window.connect("destroy", self.OnDestroy)
        window.set_default_size(240, 320)

        # - Menu
        menu = gtk.GtkMenuBar()
        fileitem = gtk.GtkMenuItem("Notepaq")
        filemenu = gtk.GtkMenu()
        fileitem.set_submenu(filemenu)
        quititem = gtk.GtkMenuItem("Quit")
        quititem.connect("activate", window.destroy)
        filemenu.add(quititem)
        menu.add(fileitem)

        # - Buttons
        new = Button("New", self.OnNewClicked, "Create")
        self.edit = Button("Edit", self.OnEditClicked)
        self.kill = Button("Destroy", self.OnDestroyClicked, "Destroy")
        quit = Button("Quit", self.OnQuitClicked)
        buttons = HButtonBox([new, self.edit, self.kill, quit])

        # - Notes list
        self.notes = gtk.GtkList()
        self.notes.connect("select-child", self.OnSelectChild)
        for note in self.mgr.notes:
            self.AddNote(note)
        bars = gtk.GtkScrolledWindow()
        bars.set_policy(gtk.POLICY_NEVER, gtk.POLICY_ALWAYS)
        bars.add_with_viewport(self.notes)

        # - Pack the window
        vbox = gtk.GtkVBox()
        vbox.pack_start(menu, 0, 0)
        vbox.pack_start(bars)
        vbox.pack_end(buttons, 0, 0)
        window.add(vbox)

        self.selected = -1
        self.SetSelected(None)
        window.show_all()
        self.window = window

    def SetSelected(self, selected):
        if selected != self.selected:
            if self.selected:
                for child in self.notes.children():
                    if child.get_data("note") == self.selected:
                        hbox = child.children()[0]
                        if len(hbox.children()) > 1:
                            hbox.remove(hbox.children()[1])
            self.selected = selected
            self.edit.set_sensitive(selected != None)
            self.kill.set_sensitive(selected != None)
            if self.selected:
                for child in self.notes.children():
                    if child.get_data("note") == self.selected:
                        hbox = child.children()[0]
                        body = self.NoteSplit(self.selected)[1]
                        if not body: body = "<no text>"
                        text = gtk.GtkLabel(body)
                        text.set_line_wrap(1)
                        frame = gtk.GtkFrame()
                        frame.set_shadow_type(gtk.SHADOW_IN)
                        frame.add(text)
                        hbox.pack_end(frame)
                        frame.show_all()

    def NoteTitle(self, note):
        return self.NoteSplit(note)[0]

    def NoteSplit(self, note):
        lines = string.split(note.text, "\n")
        return lines[0], string.join(lines[1:], "\n")
    
    def AddNote(self, note):
        title,body_text = self.NoteSplit(note)
        for child in self.notes.children()[:]:
            if child.get_data("note") == note:
                hbox = child.children()[0]
                children = hbox.children()
                label = children[0]
                if len(children) > 1:
                    body = children[1].children()[0]
                    body.set_text(body_text)
                label.set_text(title)
                return
        item = gtk.GtkListItem()
        hbox = gtk.GtkVBox()
        title = gtk.GtkLabel(title)
        hbox.pack_start(title)
        item.add(hbox)
        item.set_data("note", note)
        item.show_all()
        self.notes.add(item)

    def RemNote(self, note):
        for child in self.notes.children()[:]:
            if child.get_data("note") == note:
                self.notes.remove(child)
                child.destroy()
                note.destroy()
        self.mgr.dump()

    def Destroy(self):
        self.window.destroy()

    def OnSelectChild(self, widget, child):
        self.SetSelected(child.get_data("note"))

    def OnQuitClicked(self, widget):
        self.Destroy()

    def OnEditClicked(self, widget):
        if self.selected:
            NoteView(self.selected, self, None)

    def OnDestroyClicked(self, widget):
        if self.selected:
            self.RemNote(self.selected)
            self.SetSelected(None)

    def OnNewClicked(self, widget):
        note = Note()
        NoteView(note, self, "New Note")

    def OnDestroy(self, widget):
        self.mgr.dump_now()
        gtk.mainquit()

def Main(argv):
    if len(argv) >= 2:
        notebook = argv[1]
    else:
        #notebook = os.path.join(os.environ["HOME"], "notebook.pickle")
        notebook = os.path.join("/tmp/notebook.pickle")
    notemanager = load_note_manager(notebook)
    notemgrview = NoteManagerView(notemanager)
    gtk.mainloop()

if __name__ == "__main__": Main(sys.argv)



